/*=========================================================================
 *
 *  Copyright UMC Utrecht and contributors
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *=========================================================================*/

#ifndef __elxInstallAllComponents_h
#define __elxInstallAllComponents_h

#include "elxComponentDatabase.h"

/** The CMake generated files contain macros: */
#include "elxMacro.h"

/** The following file, generated by CMake, declares all
 * the InstallComponent functions implemented by the components. */
#include "elxInstallComponentFunctionDeclarations.h"

int
InstallAllComponents( elx::ComponentDatabase * _cdb )
{
  int ret = 0;

  /** The following file, generated by CMake, calls the
   * InstallComponent functions of the components. */

  #include "elxInstallComponentFunctionCalls.h"

  /** ret!=0 when something failed. */
  return ret;

} // end InstallAllComponents


#endif // #ifndef __elxInstallAllComponents_h
